package goblincave.gitea.nes;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.RandomAccessFile;
import java.io.Reader;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.util.LinkedHashMap;
import java.util.Set;

import javaFlacEncoder.FLAC_FileEncoder;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVRecord;
import org.apache.commons.io.FilenameUtils;

import com.diogonunes.jcolor.Ansi;
import com.diogonunes.jcolor.Attribute;

/**
 * Ridge Racer 6 Audio Extractor
 * @author Nes
 * @version 1.0, 2024-07-14
 */
public class AudioExtractor {
	
	/**
	 * Activate alternate behavior for command line terminal.
	 * Specifically prints things in a narrower width, and overwrites previous lines to add color.
	 */
	private final static boolean CMD_MODE = true;
	
	/**
	 * ANSI code to go up one line, followed by a carriage return.
	 * Useful for overwriting user input.
	 */
	private final static String ANSI_BACKLINE = "\033[F\r";
	/**
	 * ANSI code to switch to beige colored text.
	 * Useful for highlighting user input.
	 */
	private final static String ANSI_BEIGE = "\u001B[93m";
	
	/**
	 * Hex value for the ASCII sequence "RIFF".
	 * Indicates the start of a RIFF header in WAV files.
	 */
	private final static int ASCII_RIFF = 0x5249_4646;
	/**
	 * Hex value for the ASCII sequence "RIFF".
	 * Indicates the start of a RIFF header in WAV files.
	 */
	private final static int ASCII_WAVE = 0x5741_5645;
	/**
	 * Hex value for the ASCII sequence "fmt ".
	 * Indicates the start of a format chunk in WAV files.
	 */
	private final static int ASCII_fmt = 0x666D_7420;
	/**
	 * Hex value for the ASCII sequence "smpl".
	 * Indicates the start of a sample chunk in WAV files.
	 */
	private final static int ASCII_smpl = 0x736D_706C;
	/**
	 * Hex value for the ASCII sequence "data".
	 * Indicates the start of a data chunk in WAV files.
	 */
	private final static int ASCII_data = 0x6461_7461;
	/**
	 * Hex value for the ASCII sequence "ALIG".
	 * Indicates the start of an alignment chunk in WAV files.
	 */
	private final static int ASCII_ALIG = 0x414C_4947;
	/**
	 * Hex value for the ASCII sequence "x2st".
	 * Indicates the start of an XMA stream chunk in WAV files.
	 */
	private final static int ASCII_x2st = 0x7832_7374;
	/**
	 * Hex value for the ASCII sequence "seek".
	 * Indicates the start of an XMA stream chunk in WAV files.
	 */
	private final static int ASCII_seek = 0x7365_656B;
	/**
	 * Hex value for the ASCII sequence "fLaC".
	 * Indicates the start of an FLAC file.
	 */
	private final static int ASCII_fLaC = 0x664C_6143;
	
	/**
	 * BIN files must first be extracted from game disk to use this program.
	 * By default uses program directory for I/O and saves audio as WAV.<br>
	 * 
	 * You can simply run this program in the same directory as <b>pack_*.bin</b> files
	 * to launch with default arguments: <pre>extract currentDir currentDir wav</pre>
	 * 
	 * @param args - 4 optional arguments<ol>
	 * <li>"extract" or "pack" mode</li>
	 * <li>package directory containing <b>pack_*.bin</b> files</li>
	 * <li>extract directory for unpacked audio files</li>
	 * <li>extract BGM in "WAV" or "FLAC" format</li></ol>
	 * @throws URISyntaxException 
	 * @throws InterruptedException 
	 */
	public static void main(String[] args) throws URISyntaxException, InterruptedException {
		
		// Print Audio Extractor logo
		if(CMD_MODE) {
			
			// NARROW LOGO
			System.out.println(Ansi.colorize(logoFormat("Ridge", true, 8)/*.substring(0, 484) + adjustment*/, Attribute.BRIGHT_RED_TEXT()));
			Thread.sleep(500);
			
			String adjustment = "█████████   ████████████████████████████████████████████████████                     ████████████████████████";
			System.out.println(ANSI_BACKLINE + ANSI_BACKLINE + Ansi.colorize(adjustment, Attribute.BRIGHT_RED_TEXT()));
			System.out.println(Ansi.colorize(logoFormat("Racer 6", true, 9), Attribute.BRIGHT_RED_TEXT()));
			Thread.sleep(500);
			
			System.out.println(Ansi.colorize(logoFormat("Audio", true, 8), Attribute.YELLOW_TEXT()));
			Thread.sleep(500);
			
			System.out.println(Ansi.colorize(logoFormat("Xtractr", true, 8), Attribute.YELLOW_TEXT()));
			Thread.sleep(500);
			
		} else {
			
			// WIDE LOGO
			System.out.println(Ansi.colorize(logoFormat("Ridge Racer 6"), Attribute.BRIGHT_RED_TEXT()));
			Thread.sleep(500);
			
			System.out.println(Ansi.colorize(logoFormat("Audio Extractor", true, 8), Attribute.YELLOW_TEXT()));
			Thread.sleep(500);
			
		}
		
		// Print meta info
		System.out.println("Program:\tRidge Racer 6 Audio Extractor " + Ansi.colorize("v1.0", Attribute.BRIGHT_GREEN_TEXT()) + " by " 
				+ Ansi.colorize("Nes", Attribute.TEXT_COLOR(252, 42, 124)));
		System.out.println("Repository:\t" + Ansi.colorize("https://gitea.goblincave.synology.me/Nes/RR6AudioExtractor", Attribute.CYAN_TEXT()));
		System.out.println();
		
		// Initialize user input reader
		BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
		
		while(true) {
		
			// determine which operation to execute
			String operationInput = null;
			if(args.length > 0)
				operationInput = args[0];
			String operation = retrieveOperation(reader, operationInput);
			
			if(args.length == 0) { // no arguments provided
							
				switch(operation) {
					case "identify":
						File unknownFile = retrieveFile(reader, "Please enter the path of a file: " + ANSI_BEIGE);
						identify(unknownFile);
						break;
					case "extract": // extract
						// Determine package directory
						File currentDir = new File(new File(AudioExtractor.class.getProtectionDomain().getCodeSource().getLocation().toURI())
								.getParent());
						boolean packDirFound = false;
						do {
							System.out.println(Ansi.colorize("\nPackage Directory:", Attribute.BOLD(), Attribute.CYAN_TEXT()));
							System.out.println("- " + Ansi.colorize("Current", Attribute.BRIGHT_RED_TEXT()) + " directory:\t" 
									+ Ansi.colorize(currentDir.getPath(), Attribute.BRIGHT_CYAN_TEXT()));
							
						} while(!packDirFound);
						break;
					case "package": // pack
						break;
					case "patch": // patch
						File waveFile = retrieveWave(reader);
						patch(waveFile);
						break;
					case "print": // print
						Attribute color = retrieveColor(reader);
						String message = retrieveInput(reader, "Please enter your logo message: " + ANSI_BEIGE);
						System.out.println(Ansi.colorize(logoFormat(message), color));
						break;
					case "exit": // exit
						System.out.println("Program closed.");
						System.exit(0);
						break;
				}
	
			}
		
		}
		
/*
		System.exit(0);
		
		if(args.length > 0) {
			System.out.println("Arguments\t:");
			for(int i = 0; i < args.length; i++)
				System.out.print(Ansi.colorize(args[i] + ' ', Attribute.BRIGHT_GREEN_TEXT()));
		}
		
		if(args.length > 0) {
			switch(args[0]) {
				
				case "help":
					if(args.length > 1)
						switch(args[1]) {
							case "extract":
								
								break;
							case "pack":
								break;
							case "print":
								break;
						}
					System.out.println("java -jar RR6AudioExtractor.jar [operation] [package] [extract] [option]");
					break;
					
				case "extract":
					// TODO 
					break;
					
				case "pack":
					// TODO
					break;
					
			}
				
		}
		
		// Confirm input/output mode
		boolean packMode = false;
		if(args.length > 0)
			packMode = args[0].equalsIgnoreCase("pack");
		
		// Confirm package directory, else abort
		File packDirectory;
		if(args.length > 1)
			packDirectory = Path.of(args[1]).toFile();
		else packDirectory = new File(new File(AudioExtractor.class.getProtectionDomain().getCodeSource().getLocation().toURI()).getParent());
		if(packMode && !packDirectory.exists() && !packDirectory.mkdir()) {
			System.out.println("Package output directory " + packDirectory + " does not exist, and could not be created. "
					+ "Program execution aborted.");
			System.exit(1);
		} else if(!packMode && !packDirectory.exists()) {
			System.out.println("Package input directory " + packDirectory + " does not exist. Program execution aborted.");
			System.exit(1);
		}
		
		// Confirm extract directory, else abort
		File extractDirectory;
		if(args.length > 2)
			extractDirectory = Path.of(args[2]).toFile();
		else extractDirectory = packDirectory;
		if(!packMode && !extractDirectory.exists() && !extractDirectory.mkdir()) {
			System.out.println("Extract output directory " + extractDirectory + " does not exist, and could not be created. "
					+ "Program execution aborted.");
			System.exit(2);
		}
		
		// Confirm BGM output format
		boolean compressBGM = false;
		if(args.length > 3)
			compressBGM = args[3].equalsIgnoreCase("FLAC");
		
		// Begin operation
		if(packMode)
			pack(packDirectory, extractDirectory);
		else extract(packDirectory, extractDirectory, compressBGM);
*/
		
	}
	

	
	/**
	 * Helper function to prompt user to select a program task.
	 * 
	 * @param reader
	 * @param input
	 * @return operation
	 */
	private static String retrieveOperation(BufferedReader reader, String input) {
		
		String operation = null;
		String[] values = {"identify", "extract", "pack", "patch", "print", "exit"};
		
		if(input != null)
			for(int i = 0; i < values.length; i++)
				if(values[i].equalsIgnoreCase(input))
					operation = values[i];
		
		if(operation == null) {
		
			String prompt = Ansi.colorize("Operations", Attribute.BRIGHT_YELLOW_TEXT(), Attribute.BOLD()) + ":\n"
					+ "- " + Ansi.colorize("Identify", Attribute.BRIGHT_RED_TEXT()) + " a package or audio file\n"
					+ "- " + Ansi.colorize("Extract", Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F)) + " audio tracks from a package (BIN -> WAV/XMA)\n"
					+ "- " + Ansi.colorize("Convert", Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00)) + " audio tracks to another format (XMA <-> WAV <-> FLAC)\n"
					+ "- " + Ansi.colorize("Patch", Attribute.BRIGHT_GREEN_TEXT()) + " an audio track to repeat (WAV)\n"
					+ "- " + Ansi.colorize("Pack", Attribute.BRIGHT_BLUE_TEXT()) + " audio tracks into a package (WAV/XMA -> BIN)\n"
					+ "- " + Ansi.colorize("Print", Attribute.TEXT_COLOR(0x6B, 0x4C, 0xE3)) + " a Ridge Racer 6 style ASCII logo\n"
					+ "- " + Ansi.colorize("Exit", Attribute.BRIGHT_MAGENTA_TEXT()) + " program\n"
					+ "\nPlease select an operation: " + ANSI_BEIGE;
			
			operation = retrieveInput(reader, prompt, values);
			
			if(CMD_MODE) {
				
				Attribute color = null;
				switch(operation.toLowerCase()) {
					case "identify":
						color = Attribute.BRIGHT_RED_TEXT();
						break;
					case "extract":
						color = Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F);
						break;
					case "convert":
						color = Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00);
					case "patch":
						color = Attribute.BRIGHT_GREEN_TEXT();
						break;
					case "pack":
						color = Attribute.BRIGHT_BLUE_TEXT();
						break;
					case "print":
						color = Attribute.TEXT_COLOR(0x6B, 0x4C, 0xE3);
						break;
					case "exit":
						color = Attribute.BRIGHT_MAGENTA_TEXT();
						break;
				}
				System.out.println(ANSI_BACKLINE + Ansi.colorize("Please select an operation: ") + Ansi.colorize(operation, color));
				
			} else System.out.print(Ansi.colorize(""));
		
		}
		
		return operation.toLowerCase();
		
	}
	
	/**
	 * Examines a given file and reads some of its properties.
	 * 
	 * @param unknownFile
	 */
	public static void identify(File unknownFile) {
		
		// TODO check for characteristics of BIN or WAV files
		String name = unknownFile.getName();
		System.out.println("Name:\t" + name);
		
		long size = unknownFile.length();
		System.out.println("Size:\t" + size + " Bytes" + (size > 1024 ? " (" + convertBytes(size) + ")" : "") + "\n");
		
		// file extension
		int extIndex = name.lastIndexOf('.');
		String extension = null;
		if(extIndex > 0 && extIndex < name.length() - 1)
			extension = name.substring(extIndex + 1);
		
		try {
			RandomAccessFile file = new RandomAccessFile(unknownFile, "r");
			switch(extension.toLowerCase()) {
				case "bin": {
					
					break;
				}
				case "flac": {
					
					break;
				}
				case "wav":
				case "xma":
				default: {
					for(long i = 0; i < size - 4; i += 4) {
						file.seek(i);
						int value = file.readInt();
						switch(value) {
							case ASCII_RIFF: {
								readRIFFChunk(file, (int) i);
								break;
							}
							case ASCII_fmt: {
								i = readFormatChunk(file, (int) i) - 4;
								break;
							}
							case ASCII_smpl: {
								i = readSampleChunk(file, (int) i) - 4;
								break;
							}
							case ASCII_data: {
								i = readDataChunk(file, (int) i) - 4;
								break;
							}
							case ASCII_ALIG: {
								readAlignmentChunk(file, (int) i);
								break;
							}
							case ASCII_seek: {
								readx2stChunk(file, (int) i);
								break;
							}
							case ASCII_x2st: {
								readx2stChunk(file, (int) i);
								break;
							}
						}
					}
					break;
				}
			}
		} catch (IOException e) {
			System.out.println("Could not read file.");
		}
		
		// check if RIFF header
		
	}
	
	/**
	 * Helper function to read RIFF header information.
	 * 
	 * @param file
	 * @param offset
	 * @throws IOException
	 */
	public static void readRIFFChunk(RandomAccessFile file, int offset) throws IOException {

		System.out.println(formatAddress(offset, file.length()) + ":\tRIFF (Resource Interchange File Format) file");
		
		if(offset + 8 < file.length()) {
			int riffSize = readChunkSize(file, offset);
			System.out.println(formatAddress(offset + 4, file.length()) + ":\t" + riffSize + " Bytes" + (riffSize > 1024 ? " (" + convertBytes(riffSize) + ")" : ""));
		}
		
		if(offset + 12 < file.length()) {
			file.seek(offset + 8);
			if(file.readInt() == ASCII_WAVE)
				System.out.println(formatAddress(offset + 8, file.length()) + ":\tWAVE (Waveform Audio File Format) type");
			offset += 8;
		}
		
		System.out.println();
		
	}
	

	/**
	 * Helper method to read format chunk information. 
	 * 
	 * @param file
	 * @param offset from start of file to start of format chunk
	 * @return offset from start of file to end of format chunk
	 * @throws IOException
	 */
	public static int readFormatChunk(RandomAccessFile file, int offset) throws IOException {
		
		System.out.println(formatAddress(offset, file.length()) + ":\tfmt (Format) chunk");
		
		int size = 8;
		if(offset + 8 < file.length()) {
			size = readChunkSize(file, (int) offset);
			System.out.println(formatAddress(offset + 0x04, file.length()) + ":\t" + size + " Bytes" + (size > 1024 ? " (" + convertBytes(size) + ")" : ""));
		}
		
		if(offset + size < file.length()) {
			
			byte[] LEBytes = new byte[2];
			file.seek(offset + 0x08);
			file.read(LEBytes);
			int encoding = littleEndianToInt(LEBytes);
			System.out.print(formatAddress(offset + 0x08, file.length()) + ":\t");
			if(encoding == 0x0001)
				System.out.print("PCM (Pulse-Code Modulation)");
			else if(encoding == 0x0002)
				System.out.print("ADPCM (Adaptive Differential Pulse-Code Modulation)");
			else if(encoding == 0x0165)
				System.out.print("XMA (Xbox Media Audio)");
			else if(encoding == 0x0166)
				System.out.print("XMA2 (Xbox Media Audio 2)");
			else { 
				boolean found = false;
				InputStream filestream = AudioExtractor.class.getResourceAsStream("/encoding.csv");
				InputStreamReader streamReader = new InputStreamReader(filestream);
				Reader in = streamReader;
				String[] headers = { "Decimal value", "Hexadecimal value", "Description" };
				Iterable<CSVRecord> records = CSVFormat.EXCEL.builder().setHeader(headers).setSkipHeaderRecord(true).build().parse(in);
				for (CSVRecord record : records) {
				    String dec = record.get("Decimal value");
				    String desc = record.get("Description");
				    if(encoding == Integer.parseInt(dec)) {
				    	System.out.print(desc);
				    	found = true;
				    	break;
				    }
				}
				if(!found)
					System.out.print("Unknown (" + encoding + ")");
			}
			System.out.println(" encoding");
			
			if(encoding == 0x0165) {
				LEBytes = new byte[2];
				file.seek(offset + 0x10);
				file.read(LEBytes);
				int streams = littleEndianToInt(LEBytes);
				System.out.println(formatAddress(offset + 0x10, file.length()) + ":\t" + streams + (streams == 1 ? " stream" : " streams"));
				parseField(file, offset + 0x12, 1, " loop flag");
				// unknown 5 bytes ?
				parseField(file, offset + 0x18, 4, " Hz sampling rate");
				parseField(file, offset + 0x1C, 4, " loop start");
				parseField(file, offset + 0x20, 4, " loop end");
				parseField(file, offset + 0x24, 1, " loop subframe");
				// unknown 3 bytes ?
				
//				int channels = 0;
//				for(int i = 0; i < streams; i++)
//					channels += parseField(file, offset + 0x08 + 0x0C + 0x14 * i + 0x11, 1, " channel");
//				System.out.println(channels + " channels");
			} else {
				parseField(file, offset + 0x0A, 2, " channels");
				parseField(file, offset + 0x0C, 4, " Hz sampling rate");
				parseField(file, offset + 0x10, 4, " average bytes per second");
				parseField(file, offset + 0x14, 2, "-byte sample block size");
				parseField(file, offset + 0x16, 2, " significant bits per sample");
				LEBytes = new byte[2];
				file.seek(offset + 0x18);
				file.read(LEBytes);
				int extra = littleEndianToInt(LEBytes);
				if(extra != ASCII_smpl)
					parseField(file, offset + 0x18, 2, " extra format bytes");
			}
			
		}
		
		System.out.println();
		return offset + size;
			
	}
	
	private static int readDataChunk(RandomAccessFile file, int offset) throws IOException {
		
		System.out.println(formatAddress(offset, file.length()) + ":\tdata (Data) chunk");
		
		int size = 8;
		if(offset + 8 < file.length()) {
			size = readChunkSize(file, (int) offset);
			System.out.println(formatAddress(offset + 0x04, file.length()) + ":\t" + size + " Bytes" + (size > 1024 ? " (" + convertBytes(size) + ")" : ""));
		}

		System.out.println();
		return offset + size;
		
	}

	private static int readSampleChunk(RandomAccessFile file, int offset) throws IOException {
		
		System.out.println(formatAddress(offset, file.length()) + ":\tsmpl (Sample) chunk");
		
		int size = 8;
		if(offset + 8 < file.length()) {
			size = readChunkSize(file, (int) offset);
			System.out.println(formatAddress(offset + 0x04, file.length()) + ":\t" + size + " Bytes" + (size > 1024 ? " (" + convertBytes(size) + ")" : ""));
		}
		
		// 0x08 4B manufacturer code
//		parseField(file, offset + 0x08, 4, " manufacturer code");
		byte[] bytes = new byte[4];
		file.seek(offset + 0x08);
		file.read(bytes);
		int relevantBytes = bytes[0];
		String code;
		if(relevantBytes == 3)
			code = String.format("%02XH %02XH %02XH", bytes[1], bytes[2], bytes[3]);
		else code = String.format("%02XH", bytes[3]);
		
		System.out.print(formatAddress(offset + 0x08, file.length()) + ":\t");
		
		boolean found = false;
		InputStream filestream = AudioExtractor.class.getResourceAsStream("/manufacturer.csv");
		InputStreamReader streamReader = new InputStreamReader(filestream);
		Reader in = streamReader;
		String[] headers = { "SysEx ID Number", "Company Name" };
		Iterable<CSVRecord> records = CSVFormat.EXCEL.builder().setHeader(headers).setSkipHeaderRecord(true).build().parse(in);
		for (CSVRecord record : records) {
		    String ID = record.get("SysEx ID Number");
		    String manufacturer = record.get("Company Name");
		    if(code.equalsIgnoreCase(ID)) {
		    	System.out.print(manufacturer + " (" + code + ")");
		    	found = true;
		    	break;
		    }
		}
		if(!found)
			System.out.print("Unknown (" + code + ")");
		System.out.println(" manufacturer");
		
		// 0x0C 4B product code
		parseField(file, offset + 0x0C, 4, " product code");
		parseField(file, offset + 0x10, 4, "-nanosecond sample period");
		parseField(file, offset + 0x14, 4, " MIDI unity note");
		parseField(file, offset + 0x18, 4, " MIDI pitch fraction");
		parseField(file, offset + 0x1C, 4, " SMPTE format");
		parseField(file, offset + 0x20, 4, " SMPTE offset");
		int loops = parseField(file, offset + 0x24, 4, " sample loops");
		parseField(file, offset + 0x28, 4, " sample data");
		
		for(int i = 0; i < loops; i++) {
			int loopOffset = offset + 0x2C + i * 0x18;
			parseField(file, loopOffset, 4, " loop ID");
			
			bytes = new byte[4];
			file.seek(loopOffset + 0x04);
			file.read(bytes);
			int loopType = littleEndianToInt(bytes);
			System.out.print(formatAddress(loopOffset + 0x04, file.length()) + ":\t");
			String type;
			switch(loopType) {
				case 0:
					type = "Forward";
					break;
				case 1:
					type = "Alternating";
					break;
				case 2:
					type = "Backward";
					break;
				default:
					type = "Unknown (" + loopType + ")";
					break;
			}
			type += " loop type";
			System.out.println(type);
			
			file.seek(loopOffset + 0x08);
			file.read(bytes);
			int loopStart = littleEndianToInt(bytes);
			System.out.println(formatAddress(loopOffset + 0x08, file.length()) + ":\tSample # " + loopStart + " loop start");

			file.seek(loopOffset + 0x0C);
			file.read(bytes);
			int loopEnd = littleEndianToInt(bytes);
			System.out.println(formatAddress(loopOffset + 0x08, file.length()) + ":\tSample # " + loopEnd + " loop end");
			
			parseField(file, loopOffset + 0x10, 4, " loop fraction");
			parseField(file, loopOffset + 0x14, 4, " repeat count");
		}
		
		System.out.println();
		return offset + size;
		
	}
	
	public static int readSeekChunk(RandomAccessFile file, int offset) throws IOException {
		
		System.out.println(formatAddress(offset, file.length()) + ":\tseek (Seek) chunk");
		
		int size = 8;
		if(offset + 8 < file.length()) {
			size = readChunkSize(file, (int) offset);
			System.out.println(formatAddress(offset + 0x04, file.length()) + ":\t" + size + " Bytes" + (size > 1024 ? " (" + convertBytes(size) + ")" : ""));
		}
		
		return offset + size;
		
	}
	
	/**
	 * Helper method to read a little-endian byte field and print its contents.
	 * 
	 * @param file
	 * @param address
	 * @param size
	 * @param message
	 * @return value
	 * @throws IOException
	 */
	public static int parseField(RandomAccessFile file, int address, int size, String message) throws IOException {
	
		byte[] bytes = new byte[size];
		file.seek(address);
		file.read(bytes);
		int value = littleEndianToInt(bytes);
		
		System.out.println(formatAddress(address, file.length()) + ":\t" + value + message);
		
		return value;
		
	}

	private static int readx2stChunk(RandomAccessFile file, int offset) throws IOException {
		
		System.out.println(formatAddress(offset, file.length()) + ":\tx2st chunk");
		
		int size = 8;
		if(offset + 8 < file.length()) {
			size = readChunkSize(file, (int) offset);
			System.out.println(formatAddress(offset + 0x04, file.length()) + ":\t" + size + " Bytes" + (size > 1024 ? " (" + convertBytes(size) + ")" : ""));
		}
		
		System.out.println();
		return offset + size;
		
	}

	private static int readAlignmentChunk(RandomAccessFile file, int offset) throws IOException {

		System.out.println(formatAddress(offset, file.length()) + ":\tALIG (Alignment) chunk");
		
		int size = 8;
		if(offset + 8 < file.length()) {
			size = readChunkSize(file, (int) offset);
			System.out.println(formatAddress(offset + 0x04, file.length()) + ":\t" + size + " Bytes" + (size > 1024 ? " (" + convertBytes(size) + ")" : ""));
		}
		
		System.out.println();
		return offset + size;
		
	}

	private static String formatAddress(long address, long size) {
		
		int digits = (int) Math.ceil(Math.log(size) / Math.log(16));
		
		return String.format("%0" + digits + "X", address);
		
//		if(size <= 0xFFFF)
//			return String.format("%04X", address);
//		else return String.format("%08X", address);
		
	}
	
	/**
	 * Formats a file address with leading zeroes.
	 * 
	 * @param address
	 * @return padded address
	 */
	private static String formatAddress(long address) {
		
//		if(address > 0xFFFF)
			return String.format("%08X", address);
//		else return String.format("%04X", address);
		
	}

	/**
	 * Extracts audio files to the extract directory from package files located at the package directory.
	 * Also supports FLAC conversion for BGM files.
	 * 
	 * @param packDirectory
	 * @param extractDirectory
	 * @param compressBGM
	 */
	public static void extract(File packDirectory, File extractDirectory, boolean compressBGM) {
		
		// TODO Delete
		System.out.println("packDirectory:\t" + packDirectory);
		System.out.println("extractDirectory:\t" + extractDirectory);
		
		// Identify target binary files
		File[] packages = new File(packDirectory.toString()).listFiles(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.toLowerCase().endsWith(".bin");
			}
		});
		
		// Identify and extract audio within binary files
		for(int i = 0; i < packages.length; i++) {
			
			RandomAccessFile source;
			LinkedHashMap<Integer, Integer> tracklist = null;
			try {
				source = new RandomAccessFile(packages[i], "r");
				tracklist = findAudioTracks(source);
			} catch (FileNotFoundException e) {
				System.out.println("Binary file:\t" + packages[i] + "\n was not found. File skipped.");
				continue;
			} catch (IOException e) {
				System.out.println("Binary file:\t" + packages[i] + "\n could not be read. File skipped.");
				continue;
			}
			
			//TODO Delete
			System.out.println("packages[" + i + "]:\t" + packages[i]);
					
			File dir = new File(extractDirectory.getPath() + "/" + FilenameUtils.removeExtension(packages[i].getName()));
			if(!dir.exists())
				dir.mkdir();
			
			// Extract and write WAV files in directory
			Set<Integer> keys = tracklist.keySet();
			int track = 1;
			
			for(Integer key : keys) {
				
				String name = String.format("track%04d", track);
				{ // Extract WAV files
					File wav = null;
					try {
							
						// create file for storing WAV data
						wav = new File(dir.getPath() + "/" + name + ".wav");
						if(!wav.exists())
							wav.createNewFile();
					
						//TODO Delete
						System.out.println("Saving track " + String.format("%d (@0x%08X)", track, key) + " to " + wav + String.format(" (%d bytes)", tracklist.get(key)));
						
						// write selection to file
						try (
								FileInputStream inStream = new FileInputStream(packages[i]); 
								FileOutputStream outStream = new FileOutputStream(wav)
						) {
							inStream.skipNBytes(key.intValue());
							byte[] trackBytes = new byte[tracklist.get(key)];
							int readBytes = inStream.read(trackBytes);
							outStream.write(trackBytes, 0, readBytes);
						} catch (IOException e) {
							e.printStackTrace();
							System.exit(3);
						}
						
					} catch (IOException e) {
						System.out.println("An error occurred when attempting to write " + name + " to file:\t" + wav);
						e.printStackTrace();
						System.exit(3);
					}
				}
				
				track++;
				
			}
			
			if(compressBGM && packages[i].getName().equals("pack_bgm.bin")) {

				// find WAV files
				File[] WAV_Files = dir.listFiles(new FilenameFilter() {
					public boolean accept(File dir, String name) {
						return name.toLowerCase().endsWith(".wav");
					}
				});
				
				// convert WAV files to FLAC
				FLAC_FileEncoder ffe = new FLAC_FileEncoder();
								
				for(File wav : WAV_Files) {
				
					try {
						File flac = new File(dir.getPath() + "/" + FilenameUtils.removeExtension(wav.getName()) + ".flac");
						if(!flac.exists())
							flac.createNewFile();
						
						//TODO Delete
						System.out.println("Compressing WAV to FLAC:\t" + flac);
					
						ffe.encode(wav, flac);
					} catch (IOException e) {
						System.out.println("An error occurred when attempting to write file.");
						e.printStackTrace();
						System.exit(3);
					}
					
				}
				
				System.out.print("Deleting WAV files");
				// delete WAV files
				for(File wav : WAV_Files) {
					boolean deleted = false;
					while(!deleted) {
						try {
							Files.delete(wav.toPath());
							deleted = true;
						} catch(Exception e) {
							System.out.print('.');
						}
					}
				}
				
			}
			
		}
		
	}
	

	/**
	 * Packs WAV files located at the extract directory into BIN files at the package directory.
	 * 
	 * @param packDirectory
	 * @param extractDirectory
	 */
	private static void pack(File packDirectory, File extractDirectory) {
		
		// Identify extracted folders
		File[] packages = new File(extractDirectory.toString()).listFiles(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.toLowerCase().startsWith("pack_") && !name.toLowerCase().endsWith(".bin");
			}
		}); 
        
		for(int i = 0; i < packages.length; i++)
			System.out.println(packages[i].getName());
		
		for(File packDir : packages) {
			
			// 1. Compile list of WAV files in extracted directories
			File[] audioTracks = new File(packDir.toString()).listFiles(new FilenameFilter() {
				public boolean accept(File dir, String name) {
					return name.toLowerCase().endsWith(".wav");
				}
			});
			
			// 2. Check if pack files already exist in package directory
			File pack = new File(packDir.getPath() + ".bin");
			if(pack.exists());
			
			// 3. If pack file already exists, insert WAV files into existing file.
			// 4. If pack file doesn't exist, create new file then insert WAV files.
			
		}
		
		// Hypothesis 1: Does Ridge Racer 6 use a predetermined address list for reading tracks from package?
		// Test 1: Try changing the address of a track and see if the game can handle it successfully.
		
		// Hypothesis 2: Does Ridge Racer 6 loop tracks in their entirety, or do they use loop points defined in the file header?
		// Test 1: Try editing loop information and see if it affects in-game playback.
		// Test 2: Try replacing a track with a WAV file without loop data and see if it loops in in-game playback.
		
	}
	
	/**
	 * Patches a WAV file by seeking out its data and smpl chunks, 
	 * replacing any existing smpl chunk with a new smpl chunk,
	 * setting loop points at the start and end of the data chunk.
	 * 
	 * @param waveFile
	 */
	private static void patch(File waveFile) {
		
		try {
			RandomAccessFile file = new RandomAccessFile(waveFile, "r");
			
			// Also Find data location
			int dataAddress = -1;
			// Check if sample chunk already exists
			int smplAddress = -1;
			
			for(int i = 0; i < file.length(); i += 4) {
				
				file.seek(i);
				int scan = file.readInt();
				
				// match "data" chunk header
				if(scan == ASCII_data)
					dataAddress = i;
				// match "smpl" chunk header
				else if(scan == ASCII_smpl)
					smplAddress = i;
				
				// if both chunks already located, stop seeking
				if(smplAddress > -1 && dataAddress > -1)
					break;
				
			}
			
			// Determine size of data chunk
			int dataSize;
			if(dataAddress > -1)
				dataSize = readChunkSize(file, dataAddress);
			else throw new IOException("Data chunk not in file.");
			
			// Determine size of smpl chunk
			int smplSize;
			if(smplAddress > -1)
				smplSize = readChunkSize(file, smplAddress);	
			// smpl chunk may not already exist
			
			// Create new sample chunk
				// Insert 68-byte sample chunk with loop before data chunk
			
			byte[] smpl = new byte[68];
			// smpl (0)
			smpl[0] = 0x73;
			smpl[1] = 0x6D;
			smpl[2] = 0x70;
			smpl[3] = 0x6C;
			// size (4)
			smpl[4] = 0x3C;
			// manufacturer (8)
			// product (12)
			// sample period (16)
			// MIDI unity note (20)
			smpl[20] = 0x3C;
			// MIDI pitch fraction (24)
			// SMPTE Format (28)
			// SMPTE Offset (32)
			// Number of Sample Loops (36)
			smpl[36] = 0x01;
			// Sample Loops size (40)
			// loop ID (44)
			// loop type (48)
				// data start
			// loop start (52)
				// data end
			// loop end (56)
			// tuning fraction (60)
			// play count (64)
			
			
			// Set loop points to start and end of data chunk
			// Update the new file size in RIFF header
			
		} catch(IOException e) {
			System.out.println("Failed to patch file.");
		}
		
	}
	
	/**
	 * Helper function to prompt user to select a WAV file.
	 * 
	 * @param reader
	 * @return WAV file
	 */
	private static File retrieveWave(BufferedReader reader) {
		
		String prompt = "Please provide a WAV file: " + ANSI_BEIGE;
		
		File waveFile = null;
		boolean valid = false;
		do {
			
			// Retrieve an audio file
			waveFile = retrieveFile(reader, prompt, ".wav");
			
			try {
				// validate that chosen file is a proper Wave file
				RandomAccessFile file = new RandomAccessFile(waveFile, "r");
				// first 4 bytes start with RIFF
				if(file.readInt() == ASCII_RIFF) {
					// next 4 bytes contain the size of the file in little-endian representation
					file.seek(4);
					byte[] buffer = new byte[4];
					file.read(buffer); // write size bytes to buffer
					
					// calculate expected file size
					int headerSize = 8 + buffer[3] 
							+ buffer[2] * (int) Math.pow(16, 2) 
							+ buffer[1] * (int) Math.pow(16, 4)
							+ buffer[0] * (int) Math.pow(16, 6);
					
					// file is valid if actual file length matches expected size
					valid = headerSize == file.length();
					if(!valid)
						System.out.println("File length mismatch. Expected " + headerSize + " bytes, but found " + file.length() + " bytes.");
					
				} else System.out.println("Invalid RIFF header.");
				file.close();
			} catch (IOException e) {
				System.out.println("Invalid file.");
			}
			
		} while(!valid);
		
		return waveFile;
		
	}

	/**
	 * Helper function to prompt user to select a file.
	 * 
	 * @param reader
	 * @return file
	 */
	private static File retrieveFile(BufferedReader reader, String prompt) {
		
		File file = null;
		
		String input = null;
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				if(input == null)
					throw new IOException("Invalid file.");
				file = new File(input);
				if(file.exists()) {
					if(file.isFile()) {
						valid = true;
					} else System.out.println(Ansi.colorize("Path does not resolve to a file."));
				} else System.out.println(Ansi.colorize("File does not exist."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid file."));
			}
		} while(!valid);
		
		System.out.print(Ansi.colorize("\n"));
		
		return file;
		
	}
	
	/**
	 * Helper function to prompt user to select a file.
	 * 
	 * @param reader
	 * @return file
	 */
	private static File retrieveFile(BufferedReader reader, String prompt, String extension) {
		
		File file = null;
		
		String input = null;
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				file = new File(input);
				if(file.exists()) {
					if(file.getName().toLowerCase().endsWith(extension)) {
						valid = true;
					} else System.out.println(Ansi.colorize("File does not have " + extension.toUpperCase() + " file extension"));
				} else System.out.println(Ansi.colorize("File does not exist."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid file."));
			}
		} while(!valid);
		
		return file;
		
	}

	/**
	 * Helper function to prompt user to select a color.
	 * 
	 * @param reader
	 * @return color attribute
	 */
	private static Attribute retrieveColor(BufferedReader reader) {
		
		String prompt = "\nCommand parameters: " + Ansi.colorize("Print", Attribute.BRIGHT_BLUE_TEXT()) 
				+ Ansi.colorize(" [color]", Attribute.BRIGHT_YELLOW_TEXT()) + Ansi.colorize(" [message]", Attribute.BRIGHT_RED_TEXT()) 
				+ "\n" 
				+ "\n" + Ansi.colorize("Colors", Attribute.BRIGHT_YELLOW_TEXT(), Attribute.BOLD()) + ":"
				+ "\n- " + Ansi.colorize("Default", Attribute.CLEAR())
				+ "\n- " + Ansi.colorize("Black", Attribute.BLACK_TEXT()) 
						+ ", " + Ansi.colorize("Gray", Attribute.BRIGHT_BLACK_TEXT()) 
						+ ", " + Ansi.colorize("White", Attribute.BRIGHT_WHITE_TEXT())
				+ "\n- " + Ansi.colorize("Indigo", Attribute.TEXT_COLOR(0x7D, 0x1A, 0xFF)) 
						+ ", " + Ansi.colorize("Purple", Attribute.BRIGHT_MAGENTA_TEXT()) 
						+ ", " + Ansi.colorize("Violet", Attribute.TEXT_COLOR(0xEE, 0x82, 0xEE))
				+ "\n- " + Ansi.colorize("Red", Attribute.RED_TEXT()) 
						+ ", " + Ansi.colorize("Cinnabar", Attribute.BRIGHT_RED_TEXT()) 
						+ ", " + Ansi.colorize("Pink", Attribute.TEXT_COLOR(0xFF, 0xC0, 0xCB))
				+ "\n- " + Ansi.colorize("Orange", Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F)) 
						+ ", " + Ansi.colorize("Gold", Attribute.YELLOW_TEXT()) 
						+ ", " + Ansi.colorize("Yellow", Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00))
				+ "\n- " + Ansi.colorize("Green", Attribute.BRIGHT_GREEN_TEXT()) 
						+ ", " + Ansi.colorize("Chartreuse", Attribute.TEXT_COLOR(0xD1, 0xEB, 0x27)) 
						+ ", " + Ansi.colorize("Beige", Attribute.BRIGHT_YELLOW_TEXT())
				+ "\n- " + Ansi.colorize("Blue", Attribute.BRIGHT_BLUE_TEXT()) 
						+ ", " + Ansi.colorize("Teal", Attribute.TEXT_COLOR(0x00, 0x93, 0xCF)) 
						+ ", " + Ansi.colorize("Cyan", Attribute.BRIGHT_CYAN_TEXT())
				+ "\n" 
				+ "\n" + Ansi.colorize("Please select a color:") + ANSI_BEIGE + " ";
		String[] values = {"Default", "Black", "Gray", "White",	"Indigo", "Purple", "Violet", "Red", "Cinnabar", "Pink", "Orange", "Gold", 
				"Yellow", "Green", "Chartreuse", "Beige","Blue", "Teal", "Cyan"};
		
		String input = retrieveInput(reader, prompt, values);
		
		Attribute color = null;
		switch(input.toLowerCase()) {
			case "default":		color = Attribute.CLEAR(); 						break;
			case "black": 		color = Attribute.BLACK_TEXT(); 				break;
			case "gray": 		color = Attribute.BRIGHT_BLACK_TEXT(); 			break;
			case "white": 		color = Attribute.BRIGHT_WHITE_TEXT(); 			break;
			case "indigo": 		color = Attribute.TEXT_COLOR(0x7D, 0x1A, 0xFF); break;
			case "purple": 		color = Attribute.BRIGHT_MAGENTA_TEXT(); 		break;
			case "violet": 		color = Attribute.TEXT_COLOR(0xEE, 0x82, 0xEE); break;
			case "red": 		color = Attribute.RED_TEXT(); 					break;
			case "cinnabar": 	color = Attribute.BRIGHT_RED_TEXT(); 			break;
			case "pink": 		color = Attribute.TEXT_COLOR(0xFF, 0xC0, 0xCB); break;
			case "orange":		color = Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F);	break;
			case "gold":		color = Attribute.YELLOW_TEXT();				break;
			case "yellow":		color = Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00);	break;
			case "green": 		color = Attribute.BRIGHT_GREEN_TEXT(); 			break;
			case "chartreuse": 	color = Attribute.TEXT_COLOR(0xD1, 0xEB, 0x27); break;
			case "beige":		color = Attribute.BRIGHT_YELLOW_TEXT(); 		break;
			case "blue":		color = Attribute.BRIGHT_BLUE_TEXT(); 			break;
			case "teal":		color = Attribute.TEXT_COLOR(0x00, 0x93, 0xCF); break;
			case "cyan":		color = Attribute.BRIGHT_CYAN_TEXT(); 			break;
		}
		
		if(CMD_MODE)
			System.out.println(ANSI_BACKLINE + Ansi.colorize("Please select a color: ") + Ansi.colorize(input, color) + "\n");
		else System.out.println(Ansi.colorize(""));
		
		return color;
		
	}

	/**
	 * Prompts the user for input without strict validation.
	 * 
	 * @param reader
	 * @param prompt
	 * @return input
	 */
	private static String retrieveInput(BufferedReader reader, String prompt) {
		
		String input = null;
		
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				valid = true;
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid input."));
			}
		} while(!valid);
		
		return input;
		
	}
	
	/**
	 * Prompts the user for input and validates that the input matches one of the provided values.
	 * 
	 * @param reader
	 * @param prompt
	 * @param values
	 * @return input
	 */
	private static String retrieveInput(BufferedReader reader, String prompt, String[] values) {
		
		String input = null;
		
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				for(int i = 0; i < values.length; i++)
					if(input.equalsIgnoreCase(values[i])) {
						valid = true;
						break;
					}
				if(!valid)
					System.out.println(Ansi.colorize("Invalid selection."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid selection."));
			}
		} while(!valid);
		
		return input;
		
	}
	
	/**
	 * Helper function to read the size of a file chunk.
	 * 
	 * @param file
	 * @param chunkAddress
	 * @return chunk size
	 * @throws IOException 
	 */
	private static int readChunkSize(File file, int chunkAddress) throws IOException {
		
		RandomAccessFile accessedFile = new RandomAccessFile(file, "r");
		return readChunkSize(accessedFile, chunkAddress);
		
	}
	
	/**
	 * Helper function to read the size of a file chunk.
	 * 
	 * @param accessedFile
	 * @param chunkAddress
	 * @return chunk size
	 * @throws IOException 
	 */
	private static int readChunkSize(RandomAccessFile file, int chunkAddress) throws IOException {
		
		byte[] buffer = new byte[4];		
		
		file.seek(chunkAddress + 4);
		file.read(buffer);
//		System.out.println("DEBUG: " + (buffer[0] & 0xFF) + " " + (buffer[1] & 0xFF) + " " + (buffer[2] & 0xFF) + " " + (buffer[3] & 0xFF) + "");
		return 8	// size offset 
				+ (buffer[0] & 0xFF) 				// 256^0
				+ (buffer[1] & 0xFF) * 256 			// 256^1 
				+ (buffer[2] & 0xFF) * 65_536		// 256^2
				+ (buffer[3] & 0xFF) * 16_777_216;	// 256^3
		
	}
	
	/**
	 * Helper method to convert a number of bytes to a more readable form.
	 * 
	 * @param bytes
	 * @return legible form
	 */
	private static String convertBytes(long bytes) {
		
		final String[] units = new String[] {"B", "KB", "MB", "GB", "TB"};
		int i = (int) (Math.log10(bytes) / Math.log10(1024));
		double value = bytes / Math.pow(1024, i);
		return String.format("%.2f %s", value, units[i]);
		
	}

	/**
	 * Creates a map of audio tracks within provided file.
	 * 
	 * @param file
	 * @return Tracklist containing address-length pairs
	 * @throws IOException
	 */
	private static LinkedHashMap<Integer, Integer> findAudioTracks(RandomAccessFile file) throws IOException {
		
		LinkedHashMap<Integer, Integer> tracklist = new LinkedHashMap<Integer, Integer>();
		
		// Files start on addresses divisible by 0x800
		for(int a = 0; a + 0x800 < file.length(); a += 0x800) {
			
			file.seek(a);
			
			// File header starts with "RIFF" sequence
			if(file.readInt() == ASCII_RIFF) {
				
				// Next 4 bytes specify file length beyond first 8 bytes, in little-endian representation
				int length = 8;
				// Calculate remaining length of file
				for(int d = 0; d < 4; d++) {
					file.seek(a + 4 + d);
					length += file.readUnsignedByte() * ((int) Math.pow(16, 2 * d));
				}
				
				// Record file position in map
				tracklist.put(a, length);
			}
			
		}
		
		return tracklist;
		
	}

	/**
	 * Creates an ASCII logo in RR6 style using the provided text.
	 * Supports alphanumeric text with spaces only.
	 * 
	 * @param text
	 * @param italicize logo
	 * @return logo formatted text
	 */
	public static String logoFormat(String text) {
		return logoFormat(text, true, 10);
	}

	/**
	 * Creates an ASCII logo in RR6 style using the provided text.
	 * Supports alphanumeric text with spaces only.
	 * 
	 * @param string
	 * @param italicize logo
	 * @return logo formatted text
	 */
	private static String logoFormat(String string, boolean italic) {
		return logoFormat(string, italic, 10);
	}
	
	/**
	 * Creates an ASCII logo in RR6 style using the provided text.
	 * Supports alphanumeric text with spaces only.
	 * 
	 * @param text
	 * @param height 
	 * @param italicize logo
	 * @return logo formatted text
	 */
	private static String logoFormat(String text, boolean italic, int height) {
		
		// Initialize logo lines
		String[] logo = new String[10];
		for(int i = 0; i < logo.length; i++)
			logo[i] = new String();
		
		boolean spaced = false;
		for(int i = 0; i < text.length(); i++) {
			char c = text.charAt(i);

			// Add space from previous character
			if(i > 0) {
				if(c != ' ' && c!= ':' && !spaced) {
					for(int line = 0; line < logo.length - 1; line++)
						logo[line] += "   ";
					logo[9] += "███";
				}
				spaced = false;
			}
			
			// Add character to logo
			switch(c) {
				
				// Space
				case ':':
				case ' ':
					for(int line = 0; line < logo.length; line++)
						logo[line] += "    ";
					spaced = true;
					break;
					
				// Numbers
				case '0':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += " ██████████████████████    ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "████████        ████████   ";
					logo[4] += "████████        ████████   ";
					logo[5] += "████████        ████████   ";
					logo[6] += "████████        ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += " ██████████████████████    ";
					spaced = true;
					break;	
				case '1':
					logo[0] += " ███████████████   ";
					logo[1] += "████████████████   ";
					logo[2] += "        ████████   ";
					logo[3] += "        ████████   ";
					logo[4] += "        ████████   ";
					logo[5] += "        ████████   ";
					logo[6] += "        ████████   ";
					logo[7] += "        ████████   ";
					logo[8] += "        ████████   ";
					logo[9] += "        ████████   ";
					spaced = true;
					break;
				case '2':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "                ████████   ";
					logo[4] += "  ██████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "████████                   ";
					logo[7] += "████████                   ";
					logo[8] += "████████████████████████   ";
					logo[9] += "███████████████████████    ";
					spaced = true;
					break;
				case '3':
					logo[0] += " ██████████████████████    ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "                ████████   ";
					logo[4] += "        ███████████████    ";
					logo[5] += "        ██████████████     ";
					logo[6] += "                ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += " ██████████████████████    ";
					spaced = true;
					break;
				case '4':
					logo[0] += "         █████████████     ";
					logo[1] += "       ███████████████     ";
					logo[2] += "     ████       ██████     ";
					logo[3] += "   ██████       ██████     ";
					logo[4] += " ███████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "                ██████     ";
					logo[7] += "                ██████     ";
					logo[8] += "                ██████     ";
					logo[9] += "                ██████     ";
					spaced = true;
					break;
				case '5':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████                   ";
					logo[3] += "████████                   ";
					logo[4] += "████████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "                ████████   ";
					logo[7] += "                ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += "███████████████████████    ";
					spaced = true;
					break;
				case '6':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████                   ";
					logo[3] += "████████                   ";
					logo[4] += "████████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "████████        ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += "███████████████████████    ";
					spaced = true;
					break;
				case '7':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "                ████████   ";
					logo[4] += "                ████████   ";
					logo[5] += "                ████████   ";
					logo[6] += "                ████████   ";
					logo[7] += "                ████████   ";
					logo[8] += "                ████████   ";
					logo[9] += "                ████████   ";
					spaced = true;
					break;
				case '8':
					logo[0] += " ██████████████████████    ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "████████        ████████   ";
					logo[4] += "  █████████████████████    ";
					logo[5] += " █████████████████████     ";
					logo[6] += "████████        ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += " ██████████████████████    ";
					spaced = true;
					break;
				case '9':
					logo[0] += " ██████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "████████        ████████   ";
					logo[4] += " ███████████████████████    ";
					logo[5] += "   █████████████████████     ";
					logo[6] += "                ████████   ";
					logo[7] += "                ████████   ";
					logo[8] += "                ████████   ";
					logo[9] += "                ████████    ";
					spaced = true;
					break;
				
				// Capital letters
				case 'A':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'B':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "███████████ ";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'C':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'D':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'E':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'F':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'G':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████  ██████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'H':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████████████";
					logo[6] += "████    ████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'I':
					logo[0] += "    ";
					logo[1] += "    ";
					logo[2] += "████";
					logo[3] += "████";
					logo[4] += "████";
					logo[5] += "████";
					logo[6] += "████";
					logo[7] += "████";
					logo[8] += "    ";
					logo[9] += "████";
					break;
				case 'J':
					logo[0] += "        ";
					logo[1] += "        ";
					logo[2] += "  ██████";
					logo[3] += "    ████";
					logo[4] += "    ████";
					logo[5] += "    ████";
					logo[6] += "    ████";
					logo[7] += "████████";
					logo[8] += "        ";
					logo[9] += "████████";
					break;
				case 'K':
					logo[0] += "             ";
					logo[1] += "             ";
					logo[2] += "████    ████ ";
					logo[3] += "████   ████  ";
					logo[4] += "████  ████   ";
					logo[5] += "████ ████    ";
					logo[6] += "████████     ";
					logo[7] += "████  █████  ";
					logo[8] += "        █████";
					logo[9] += "███████   ███";
					spaced = true;
					break;
				case 'L':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████        ";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'M':
					logo[0] += "              ";
					logo[1] += "              ";
					logo[2] += "████      ████";
					logo[3] += "██████  ██████";
					logo[4] += "████ ████ ████";
					logo[5] += "████      ████";
					logo[6] += "████      ████";
					logo[7] += "████      ████";
					logo[8] += "              ";
					logo[9] += "██████████████";
					break;
				case 'N':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += "████    ████   ";
					logo[3] += "██████  ████   ";
					logo[4] += "███████ ████   ";
					logo[5] += "████ ███████   ";
					logo[6] += "████  ██████   ";
					logo[7] += "████    ████   ";
					logo[8] += "          ████ ";
					logo[9] += "█████████   ███";
					spaced = true;
					break;
				case 'O':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += " ██████████ ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += " ██████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'P':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'Q':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += " ███████████   ";
					logo[3] += "████    ████   ";
					logo[4] += "████    ████   ";
					logo[5] += "████    ████   ";
					logo[6] += "████    ████   ";
					logo[7] += "████████████   ";
					logo[8] += "          ████ ";
					logo[9] += "█████████   ███";
					spaced = true;
					break;
				case 'R':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += "████████████   ";
					logo[3] += "████    ████   ";
					logo[4] += "████    ████   ";
					logo[5] += "████    ████   ";
					logo[6] += "███████████    ";
					logo[7] += "█████   █████  ";
					logo[8] += "          █████";
					logo[9] += "█████████   ███";
					spaced = true;
					break;
				case 'S':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "  ████      ";
					logo[5] += "      ████  ";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'T':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "    ████    ";
					logo[4] += "    ████    ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'U':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'V':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "  ████████  ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'W':
					logo[0] += "              ";
					logo[1] += "              ";
					logo[2] += "████      ████";
					logo[3] += "████      ████";
					logo[4] += "████      ████";
					logo[5] += "████ ████ ████";
					logo[6] += "██████  ██████";
					logo[7] += "████      ████";
					logo[8] += "              ";
					logo[9] += "██████████████";
					break;
				case 'X':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += "████    ████   ";
					logo[3] += "████    ████   ";
					logo[4] += " ████  ████    ";
					logo[5] += "    ████       ";
					logo[6] += " ████  ████    ";
					logo[7] += "████    ████   ";
					logo[8] += "          █████";
					logo[9] += "█████████   ███";
					spaced = true;
					break;
				case 'Y':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += " ████  ████ ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'Z':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "        ████";
					logo[4] += "      ████  ";
					logo[5] += "  ████      ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
					
				// Lowercase letters
				case 'a':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'b':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "██████████  ";
					logo[6] += "████    ████";
					logo[7] += "███████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'c':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'd':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "███████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'e':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'f':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'g':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "███████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'h':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'i':
					logo[0] += "    ";
					logo[1] += "    ";
					logo[2] += "    ";
					logo[3] += "████";
					logo[4] += "████";
					logo[5] += "████";
					logo[6] += "████";
					logo[7] += "████";
					logo[8] += "    ";
					logo[9] += "████";
					break;
				case 'j':
					logo[0] += "        ";
					logo[1] += "        ";
					logo[2] += "        ";
					logo[3] += "  ██████";
					logo[4] += "    ████";
					logo[5] += "    ████";
					logo[6] += "    ████";
					logo[7] += "████████";
					logo[8] += "    ";
					logo[9] += "████████";
					break;
				case 'k':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████  ████  ";
					logo[5] += "████████    ";
					logo[6] += "████  ████  ";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'l':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'm':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████   █████";
					logo[4] += "█████ ██████";
					logo[5] += "████ ██ ████";
					logo[6] += "████    ████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'n':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "█████   ████";
					logo[5] += "████ ██ ████";
					logo[6] += "████   █████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'o':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ██████████ ";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += " ██████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'p':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "██████████  ";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'q':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "  ██████████";
					logo[7] += "        ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'r':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "██████████  ";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 's':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████        ";
					logo[5] += "    ████    ";
					logo[6] += "        ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 't':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "    ████    ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'u':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'v':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "  ████  ████";
					logo[7] += "     █████  ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'w':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += "               ";
					logo[3] += "███   ███   ███";
					logo[4] += "███   ███   ███";
					logo[5] += "███   ███   ███";
					logo[6] += "  ██  ████  ███";
					logo[7] += "    ████  ████ ";
					logo[8] += "               ";
					logo[9] += "███████████████";
					break;
				case 'x':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "  ████ ████ ";
					logo[5] += "    ████    ";
					logo[6] += " ████ ████  ";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'y':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "  ████ ████ ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'z':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "      ████  ";
					logo[5] += "    ████    ";
					logo[6] += "  ████      ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				default:
					System.out.println("Unable to convert character '" + c + "'");
					break;
			}
		}
		
		// Italicize logo
		if(italic) {
			for(int i = 0; i < logo.length; i++) {
				// i = 0 -> 9
				for(int x = logo.length - (i + 1); x > 0; x--)
					// x = 10 - (i + 1)  
					logo[i] = " " + logo[i]; // line 0 gets 10 spaces, line 1 gets 9 spaces... line 9 gets 0 spaces
			}
		}
		
		// Compile logo into single text element
		StringBuilder logoText = new StringBuilder();
		for(int i = logo.length - height; i < logo.length; i++) {
			logoText.append(logo[i] + "\n");
		}
		
		return logoText.toString();
		
	}
	
	/**
	 * Helper method that converts a little-endian byte array into a long value.
	 * 
	 * @param bytes (up to 8 bytes)
	 * @return value
	 */
	public static long littleEndianToLong(byte[] bytes) {
		if(bytes.length > 8)
			throw new IllegalArgumentException("Byte array must be 8 bytes or shorter.");
		long value = 0;
		for(int i = 0; i < bytes.length; i++)
			value |= (long) (bytes[i] & 0xFF) << (8 * i);
		return value;
	}
	
	/**
	 * Helper method that converts a little-endian byte array into a long value.
	 * 
	 * @param bytes (up to 8 bytes)
	 * @return value
	 */
	public static int littleEndianToInt(byte[] bytes) {
		if(bytes.length > 8)
			throw new IllegalArgumentException("Byte array must be 4 bytes or shorter.");
		int value = 0;
		for(int i = 0; i < bytes.length; i++)
			value |= (int) (bytes[i] & 0xFF) << (8 * i);
		return value;
	}
	
}
